@@ -1,7 +1,9 @@
-from typing import Dict
+from typing import Optional, Dict
 import time
 import random
 from enum import Enum
+
+from .interface_network import NetworkProtocol
 from .statsig_options import StatsigOptions
 
 
@@ -14,6 +16,7 @@ class Context(Enum):
 
 class Key(Enum):
     DOWNLOAD_CONFIG_SPECS = "download_config_specs"
+    DATA_STORE_CONFIG_SPECS = "data_store_config_specs"
     BOOTSTRAP = "bootstrap"
     OVERALL = "overall"
     GET_ID_LIST = "get_id_list"
@@ -63,22 +66,24 @@ class Marker:
 
     def __init__(
         self,
-        key: Key = None,
-        action: Action = None,
-        timestamp: float = None,
-        step: Step = None,
-        statusCode: int = None,
-        success: bool = None,
-        url: str = None,
-        idListCount: int = None,
-        reason: str = None,
-        sdkRegion: str = None,
-        markerID: str = None,
-        attempt: int = None,
-        retryLimit: int = None,
-        isRetry: bool = None,
-        configName: str = None,
-        error: dict = None,
+        key: Optional[Key] = None,
+        action: Optional[Action] = None,
+        timestamp: Optional[float] = None,
+        step: Optional[Step] = None,
+        statusCode: Optional[int] = None,
+        success: Optional[bool] = None,
+        url: Optional[str] = None,
+        idListCount: Optional[int] = None,
+        reason: Optional[str] = None,
+        sdkRegion: Optional[str] = None,
+        markerID: Optional[str] = None,
+        attempt: Optional[int] = None,
+        retryLimit: Optional[int] = None,
+        isRetry: Optional[bool] = None,
+        configName: Optional[str] = None,
+        error: Optional[dict] = None,
+        payloadSize: Optional[int] = None,
+        networkProtocol: Optional[NetworkProtocol] = None,
     ):
         self.key = key
         self.action = action
@@ -96,6 +101,8 @@ class Marker:
         self.isRetry = isRetry
         self.configName = configName
         self.error = error
+        self.payloadSize = payloadSize
+        self.networkProtocol = networkProtocol
 
     def to_dict(self) -> Dict:
         marker_dict = {
@@ -115,6 +122,8 @@ class Marker:
             "isRetry": self.isRetry,
             "configName": self.configName,
             "error": self.error,
+            "payloadSize": self.payloadSize,
+            "networkProtocol": self.networkProtocol.value if self.networkProtocol is not None else None,
         }
         return {k: v for k, v in marker_dict.items() if v is not None}
 
@@ -125,14 +134,14 @@ class Marker:
         if data is not None:
             for key, value in data.items():
                 setattr(self, key, value)
-        return Diagnostics.add_marker(self)
+        return self
 
     def end(self, data=None):
         self.action = Action.END
         if data is not None:
             for key, value in data.items():
                 setattr(self, key, value)
-        return Diagnostics.add_marker(self)
+        return self
 
     # Select Step #
 
@@ -150,13 +159,17 @@ class Marker:
         self.key = Key.DOWNLOAD_CONFIG_SPECS
         return self
 
+    def data_store_config_specs(self):
+        self.key = Key.DATA_STORE_CONFIG_SPECS
+        return self
+
     def bootstrap(self):
         self.key = Key.BOOTSTRAP
         return self
 
     def overall(self):
         self.context = (
-            Context.INITIALIZE.value
+            Context.INITIALIZE
         )  # overall is only ever used in initialize
         self.key = Key.OVERALL
         return self
@@ -170,23 +183,27 @@ class Marker:
         return self
 
     def api_call(self, key: Key):
-        self.context = Context.API_CALL.value
+        self.context = Context.API_CALL
         self.key = key
         return self
 
+    def log_event(self):
+        self.context = Context.LOG_EVENT
+        return self
 
-class DiagnosticsImpl:
+
+class Diagnostics:
     def __init__(self, markers=None):
         self.context_to_markers = markers or {
-            "initialize": [],
-            "config_sync": [],
-            "event_logging": [],
-            "api_call": [],
+            Context.INITIALIZE: [],
+            Context.CONFIG_SYNC: [],
+            Context.LOG_EVENT: [],
+            Context.API_CALL: [],
         }
-        self.context = "initialize"
+        self.context = Context.INITIALIZE
         self.default_max_markers = 50
         self.maxMarkers = {
-            context.value: self.default_max_markers for context in Context
+            context: self.default_max_markers for context in Context
         }
 
         self.sampling_rate = {
@@ -207,16 +224,13 @@ class DiagnosticsImpl:
         self.logger = logger
         return self
 
-    def mark(self):
-        return Marker()
-
     def set_context(self, context: Context):
         self.context = context
         return self
 
     def add_marker(self, marker):
         context = marker.context if marker.context is not None else self.context
-        if self.disabled and context == Context.API_CALL.value:
+        if self.disabled and context == Context.API_CALL:
             return False
         max_markers = self.maxMarkers.get(context, self.default_max_markers)
         cur_length = len(self.context_to_markers[context])
@@ -237,11 +251,8 @@ class DiagnosticsImpl:
     def clear_context(self, context):
         self.context_to_markers[context] = []
 
-    def log_diagnostics(self, context: Context, key: Key = None):
-        if (
-            self.logger is None
-            or len(self.context_to_markers[context]) == 0
-        ):
+    def log_diagnostics(self, context: Context, key: Optional[Key] = None):
+        if self.logger is None or len(self.context_to_markers[context]) == 0:
             return
 
         metadata = {
@@ -250,7 +261,7 @@ class DiagnosticsImpl:
             ],
             "context": context,
         }
-        if context == Context.INITIALIZE.value:
+        if context == Context.INITIALIZE:
             metadata["statsigOptions"] = (
                 self.statsig_options.get_logging_copy()
                 if isinstance(self.statsig_options, StatsigOptions)
@@ -273,7 +284,7 @@ class DiagnosticsImpl:
             elif value > MAX_SAMPLING_RATE:
                 self.sampling_rate[key] = MAX_SAMPLING_RATE
             else:
-                self.sampling_rate[key] = value
+                self.sampling_rate[key] = int(value)
 
         for samplingRateKey in SamplingRate:
             safe_set(samplingRateKey.value, obj.get(samplingRateKey.value))
@@ -281,12 +292,11 @@ class DiagnosticsImpl:
     def set_statsig_options(self, options: StatsigOptions):
         self.statsig_options = options
 
-    def should_log_diagnostics(self, context: Context, key: Key = None) -> bool:
+    def should_log_diagnostics(self, context: Context, key: Optional[Key] = None) -> bool:
         rand = random.random() * MAX_SAMPLING_RATE
-
-        if context == Context.LOG_EVENT.value:
+        if context == Context.LOG_EVENT:
             return rand < self.sampling_rate.get(SamplingRate.LOG_EVENT.value, 0)
-        if context == Context.INITIALIZE.value:
+        if context == Context.INITIALIZE:
             return rand < self.sampling_rate.get(SamplingRate.INITIALIZE.value, 0)
         if context == Context.API_CALL:
             return rand < self.sampling_rate.get(SamplingRate.API_CALL.value, 0)
@@ -296,39 +306,11 @@ class DiagnosticsImpl:
             return rand < self.sampling_rate.get(SamplingRate.DCS.value, 0)
         return rand < DEFAULT_SAMPLING_RATE  # error in code
 
-
-class Diagnostics:
-    instance = None
-
     @staticmethod
-    def initialize():
-        Diagnostics.instance = DiagnosticsImpl()
-        Diagnostics.set_diagnostics_enabled = (
-            Diagnostics.instance.set_diagnostics_enabled
-        )
-        Diagnostics.set_logger = Diagnostics.instance.set_logger
-        Diagnostics.mark = Diagnostics.instance.mark
-        Diagnostics.add_marker = Diagnostics.instance.add_marker
-        Diagnostics.get_marker_count = Diagnostics.instance.get_marker_count
-        Diagnostics.get_markers = Diagnostics.instance.get_markers
-        Diagnostics.set_max_markers = Diagnostics.instance.set_max_markers
-        Diagnostics.set_context = Diagnostics.instance.set_context
-        Diagnostics.clear_context = Diagnostics.instance.clear_context
-        Diagnostics.log_diagnostics = Diagnostics.instance.log_diagnostics
-        Diagnostics.set_sampling_rate = Diagnostics.instance.set_sampling_rate
-        Diagnostics.set_statsig_options = Diagnostics.instance.set_statsig_options
-        Diagnostics.should_log_diagnostics = Diagnostics.instance.should_log_diagnostics
-
-    @staticmethod
-    def format_error(e: Exception):
+    def format_error(e: Optional[Exception]):
         if e is None:
             return None
         return {
             "name": type(e).__name__,
+            "message": str(e)
         }
-
-    @staticmethod
-    def _safe_get_field(data: dict, field: str) -> str:
-        if field in data:
-            return str(data[field])
-        return None
