@@ -1,17 +1,20 @@
 import base64
+import re
 import time
 from datetime import datetime
-import re
 from hashlib import sha256
-from struct import unpack
-from ua_parser import user_agent_parser
-from ip3country import CountryLookup
+from typing import Dict
 
-from .statsig_user import StatsigUser
+from ip3country import CountryLookup
+from ua_parser import user_agent_parser
+
+from . import globals
 from .client_initialize_formatter import ClientInitializeResponseFormatter
+from .config_evaluation import _ConfigEvaluation
 from .evaluation_details import EvaluationDetails, EvaluationReason
 from .spec_store import _SpecStore
-from .config_evaluation import _ConfigEvaluation
+from .statsig_user import StatsigUser
+from .utils import HashingAlgorithm, sha256_hash
 
 
 class _Evaluator:
@@ -19,9 +22,9 @@ class _Evaluator:
         self._spec_store = spec_store
 
         self._country_lookup = CountryLookup()
-        self._gate_overrides = {}
-        self._config_overrides = {}
-        self._layer_overrides = {}
+        self._gate_overrides: Dict[str, dict] = {}
+        self._config_overrides: Dict[str, dict] = {}
+        self._layer_overrides: Dict[str, dict] = {}
 
     def override_gate(self, gate, value, user_id=None):
         gate_overrides = self._gate_overrides.get(gate)
@@ -73,19 +76,41 @@ class _Evaluator:
         self._config_overrides = {}
         self._layer_overrides = {}
 
-    def get_client_initialize_response(self, user: StatsigUser, client_sdk_key=None):
+    def clean_exposures(self, exposures):
+        seen: Dict[str, bool] = {}
+        result = []
+        for exposure in exposures:
+            if exposure['gate'].startswith('segment:'):
+                continue
+            key = f"{exposure['gate']}|{exposure['gateValue']}|{exposure['ruleID']}"
+            if not seen.get(key, False):
+                seen[key] = True
+                result.append(exposure)
+        return result
+
+    def get_client_initialize_response(
+            self,
+            user: StatsigUser,
+            hash: HashingAlgorithm,
+            client_sdk_key=None,
+            include_local_override=False,
+    ):
         if not self._spec_store.is_ready_for_checks():
             return None
 
+        if self._spec_store.last_update_time() == 0:
+            return None
+
         return ClientInitializeResponseFormatter \
-            .get_formatted_response(self.__eval_config, user, self._spec_store, client_sdk_key)
+            .get_formatted_response(self.__eval_config, user, self._spec_store, self, hash, client_sdk_key,
+                                    include_local_override)
 
     def _create_evaluation_details(self, reason: EvaluationReason):
         if reason == EvaluationReason.uninitialized:
             return EvaluationDetails(0, 0, reason)
 
         return EvaluationDetails(
-            self._spec_store.last_update_time, self._spec_store.initial_update_time, reason)
+            self._spec_store.last_update_time(), self._spec_store.initial_update_time, reason)
 
     def __lookup_gate_override(self, user, gate):
         gate_overrides = self._gate_overrides.get(gate)
@@ -94,7 +119,7 @@ class _Evaluator:
 
         eval_details = self._create_evaluation_details(
             EvaluationReason.local_override)
-        override = gate_overrides.get(user.user_id)
+        override = self.__lookup_override(gate_overrides, user)
         if override is not None:
             return _ConfigEvaluation(boolean_value=override, rule_id="override",
                                      evaluation_details=eval_details)
@@ -106,6 +131,9 @@ class _Evaluator:
 
         return None
 
+    def lookup_gate_override(self, user, gate):
+        return self.__lookup_gate_override(user, gate)
+
     def __lookup_config_override(self, user, config):
         config_overrides = self._config_overrides.get(config)
         if config_overrides is None:
@@ -113,7 +141,7 @@ class _Evaluator:
 
         eval_details = self._create_evaluation_details(
             EvaluationReason.local_override)
-        override = config_overrides.get(user.user_id)
+        override = self.__lookup_override(config_overrides, user)
         if override is not None:
             return _ConfigEvaluation(json_value=override, rule_id="override",
                                      evaluation_details=eval_details)
@@ -124,6 +152,9 @@ class _Evaluator:
                 json_value=all_override, rule_id="override", evaluation_details=eval_details)
         return None
 
+    def lookup_config_override(self, user, config):
+        return self.__lookup_config_override(user, config)
+
     def __lookup_layer_override(self, user, config):
         layer_overrides = self._layer_overrides.get(config)
         if layer_overrides is None:
@@ -131,7 +162,7 @@ class _Evaluator:
 
         eval_details = self._create_evaluation_details(
             EvaluationReason.local_override)
-        override = layer_overrides.get(user.user_id)
+        override = self.__lookup_override(layer_overrides, user)
         if override is not None:
             return _ConfigEvaluation(json_value=override, rule_id="override",
                                      evaluation_details=eval_details)
@@ -142,7 +173,25 @@ class _Evaluator:
                 json_value=all_override, rule_id="override", evaluation_details=eval_details)
         return None
 
-    def check_gate(self, user, gate):
+    def __lookup_override(self, config_overrides, user):
+        override = config_overrides.get(user.user_id)
+        if override is None and user.custom_ids is not None:
+            for id_name in user.custom_ids:
+                override = config_overrides.get(user.custom_ids[id_name])
+                if override is not None:
+                    break
+        return override
+
+    def unsupported_or_unrecognized(self, config_name):
+        if config_name in self._spec_store.unsupported_configs:
+            return _ConfigEvaluation(
+                evaluation_details=self._create_evaluation_details(
+                    EvaluationReason.unsupported))
+        return _ConfigEvaluation(
+            evaluation_details=self._create_evaluation_details(
+                EvaluationReason.unrecognized))
+
+    def check_gate(self, user, gate, end_result=None, is_nested=False):
         override = self.__lookup_gate_override(user, gate)
         if override is not None:
             return override
@@ -151,12 +200,17 @@ class _Evaluator:
             return _ConfigEvaluation(
                 evaluation_details=self._create_evaluation_details(
                     EvaluationReason.uninitialized))
-
         eval_gate = self._spec_store.get_gate(gate)
-        return self.__eval_config(user, eval_gate)
+        if eval_gate is None:
+            globals.logger.debug(f"Gate {gate} not found in the store. Are you sure the gate name is correct?")
+            return self.unsupported_or_unrecognized(gate)
+        if end_result is None:
+            end_result = _ConfigEvaluation()
+        self.__eval_config(user, eval_gate, end_result, is_nested)
+        return end_result
 
-    def get_config(self, user, config):
-        override = self.__lookup_config_override(user, config)
+    def get_config(self, user, config_name, is_exp=False):
+        override = self.__lookup_config_override(user, config_name)
         if override is not None:
             return override
 
@@ -165,11 +219,18 @@ class _Evaluator:
                 evaluation_details=self._create_evaluation_details(
                     EvaluationReason.uninitialized))
 
-        eval_config = self._spec_store.get_config(config)
-        return self.__eval_config(user, eval_config)
+        eval_config = self._spec_store.get_config(config_name)
+        if eval_config is None:
+            globals.logger.debug(
+                f"{'Config' if is_exp else 'Experiment'} {config_name} not found in the store. Are you sure the config name is correct?"
+            )
+            return self.unsupported_or_unrecognized(config_name)
+        result = _ConfigEvaluation()
+        self.__eval_config(user, eval_config, result)
+        return result
 
-    def get_layer(self, user, layer):
-        override = self.__lookup_layer_override(user, layer)
+    def get_layer(self, user, layer_name):
+        override = self.__lookup_layer_override(user, layer_name)
         if override is not None:
             return override
 
@@ -178,15 +239,29 @@ class _Evaluator:
                 evaluation_details=self._create_evaluation_details(
                     EvaluationReason.uninitialized))
 
-        eval_layer = self._spec_store.get_layer(layer)
-        return self.__eval_config(user, eval_layer)
+        eval_layer = self._spec_store.get_layer(layer_name)
+        if eval_layer is None:
+            globals.logger.debug(f"Layer {layer_name} not found in the store. Are you sure the layer name is correct?")
+            return self.unsupported_or_unrecognized(layer_name)
+        result = _ConfigEvaluation()
+        self.__eval_config(user, eval_layer, result)
+        return result
 
-    def __eval_config(self, user, config):
+    def __eval_config(self, user, config, end_result, is_nested=False):
         if config is None:
-            return _ConfigEvaluation(evaluation_details=self._create_evaluation_details(
-                EvaluationReason.unrecognized))
-
-        return self.__evaluate(user, config)
+            end_result.evaluation_details = self._create_evaluation_details(
+                EvaluationReason.unrecognized)
+            return
+        try:
+            self.__evaluate(user, config, end_result, is_nested)
+            end_result.evaluation_details = self._create_evaluation_details(
+                self._spec_store.init_reason)
+        except RecursionError:
+            raise
+        except Exception:
+            end_result.evaluation_details = self._create_evaluation_details(
+                EvaluationReason.error)
+            end_result.rule_id = "error"
 
     def __check_id_in_list(self, id, list_name):
         curr_list = self._spec_store.get_id_list(list_name)
@@ -197,67 +272,58 @@ class _Evaluator:
             sha256(str(id).encode('utf-8')).digest()).decode('utf-8')[0:8]
         return hashed in ids
 
-    def __evaluate(self, user, config):
-        exposures = []
-        enabled = config.get("enabled", False)
-        default_value = config.get("defaultValue", {})
-        evaluation_details = self._create_evaluation_details(
-            self._spec_store.init_reason)
-        if not enabled:
-            return _ConfigEvaluation(False, False, default_value, "disabled", exposures,
-                                     evaluation_details=evaluation_details)
+    def __evaluate(self, user, config, end_result, is_nested=False):
+        if not config.get("enabled", False):
+            self.__finalize_eval_result(config, end_result, False, None, is_nested)
+            return
 
         for rule in config.get("rules", []):
-            result = self.__evaluate_rule(user, rule)
-            if result.unsupported:
-                return _ConfigEvaluation(True, False, default_value, "", exposures,
-                                         evaluation_details=self._create_evaluation_details(
-                                             EvaluationReason.unsupported))
-            if result.secondary_exposures is not None and len(
-                    result.secondary_exposures) > 0:
-                exposures = exposures + result.secondary_exposures
-            if result.boolean_value:
-                delegated_result = self.__evaluate_delegate(
-                    user, rule, exposures)
-                if delegated_result is not None:
-                    return delegated_result
+            self.__evaluate_rule(user, rule, end_result)
+            if end_result.boolean_value:
+                if self.__evaluate_delegate(user, rule, end_result) is not None:
+                    self.__finalize_exposures(end_result)
+                    return
 
                 user_passes = self.__eval_pass_percentage(user, rule, config)
-                return _ConfigEvaluation(
-                    False,
-                    user_passes,
-                    result.json_value if user_passes else default_value,
-                    result.rule_id,
-                    exposures,
-                    is_experiment_group=result.is_experiment_group,
-                    evaluation_details=evaluation_details,
-                    group_name=result.group_name
-                )
+                self.__finalize_eval_result(config, end_result, user_passes, rule, is_nested)
+                return
 
-        return _ConfigEvaluation(False, False, default_value, "default", exposures,
-                                 evaluation_details=evaluation_details)
+        self.__finalize_eval_result(config, end_result, False, None, is_nested)
 
-    def __evaluate_rule(self, user, rule):
-        exposures = []
-        eval_result = True
+    def __finalize_eval_result(self, config, end_result, did_pass, rule, is_nested=False):
+        end_result.boolean_value = did_pass
+        if config.get("forwardAllExposures", False):
+            end_result.forward_all_exposures = True
+
+        if rule is None:
+            end_result.json_value = config.get("defaultValue", {})
+            end_result.group_name = None
+            end_result.is_experiment_group = False
+            end_result.rule_id = "default" if config.get("enabled") else "disabled"
+        else:
+            end_result.json_value = rule.get("returnValue") if did_pass else config.get("defaultValue", {})
+            end_result.group_name = rule.get("groupName", None)
+            end_result.is_experiment_group = rule.get("isExperimentGroup", False)
+            end_result.rule_id = rule.get("id", "")
+            end_result.sample_rate = rule.get("samplingRate", None)
+
+        if not is_nested:
+            self.__finalize_exposures(end_result)
+
+    def __finalize_exposures(self, end_result):
+        end_result.secondary_exposures = self.clean_exposures(end_result.secondary_exposures)
+        end_result.undelegated_secondary_exposures = self.clean_exposures(end_result.undelegated_secondary_exposures)
+
+    def __evaluate_rule(self, user, rule, end_result):
+        total_eval_result = True
         for condition in rule.get("conditions", []):
-            result = self.__evaluate_condition(user, condition)
-            if result.unsupported:
-                return result
-            if result.secondary_exposures is not None and len(
-                    result.secondary_exposures) > 0:
-                exposures = exposures + result.secondary_exposures
-            if not result.boolean_value:
-                eval_result = False
-        return_value = rule.get("returnValue", {})
-        rule_id = rule.get("id", "")
+            eval_result = self.__evaluate_condition(user, condition, end_result)
+            if not eval_result:
+                total_eval_result = False
+        end_result.boolean_value = total_eval_result
 
-        return _ConfigEvaluation(False, eval_result, return_value, rule_id, exposures,
-                                 is_experiment_group=rule.get("isExperimentGroup", False),
-                                 group_name=rule.get("groupName", None))
-
-    def __evaluate_delegate(self, user, rule, exposures):
-        config_delegate = rule.get('configDelegate', None)
+    def __evaluate_delegate(self, user, rule, end_result):
+        config_delegate = rule.get("configDelegate", None)
         if config_delegate is None:
             return None
 
@@ -265,62 +331,53 @@ class _Evaluator:
         if config is None:
             return None
 
-        delegated_result = self.__evaluate(user, config)
-        delegated_result.explicit_parameters = config.get(
-            "explicitParameters", [])
-        delegated_result.allocated_experiment = config_delegate
-        delegated_result.secondary_exposures = exposures + \
-                                               delegated_result.secondary_exposures
-        delegated_result.undelegated_secondary_exposures = exposures
-        return delegated_result
+        end_result.undelegated_secondary_exposures = end_result.secondary_exposures[:]
 
-    def __evaluate_condition(self, user, condition):
+        self.__evaluate(user, config, end_result, True)
+        end_result.explicit_parameters = config.get(
+            "explicitParameters", [])
+        end_result.allocated_experiment = config_delegate
+        return end_result
+
+    def __evaluate_condition(self, user, condition, end_result):
         value = None
         type = condition.get("type", "").upper()
         target = condition.get("targetValue")
         field = condition.get("field", "")
         id_Type = condition.get("idType", "userID")
         if type == "PUBLIC":
-            return _ConfigEvaluation(False, True)
+            return True
         if type in ("FAIL_GATE", "PASS_GATE"):
-            other_result = self.check_gate(user, target)
-            if other_result.unsupported:
-                return _ConfigEvaluation(True)
+            self.check_gate(user, target, end_result, True)
+
             new_exposure = {
                 "gate": target,
-                "gateValue": "true" if other_result.boolean_value else "false",
-                "ruleID": other_result.rule_id
+                "gateValue": "true" if end_result.boolean_value else "false",
+                "ruleID": end_result.rule_id
             }
-            exposures = [new_exposure]
-            if other_result.secondary_exposures is not None and len(
-                    other_result.secondary_exposures) > 0:
-                exposures = other_result.secondary_exposures + exposures
-            pass_gate = other_result.boolean_value if type == "PASS_GATE" else not other_result.boolean_value
-            return _ConfigEvaluation(
-                False, pass_gate, {}, "", exposures)
+
+            end_result.secondary_exposures.append(new_exposure)
+
+            pass_gate = end_result.boolean_value if type == "PASS_GATE" else not end_result.boolean_value
+            return pass_gate
         if type in ("MULTI_PASS_GATE", "MULTI_FAIL_GATE"):
             if target is None or len(target) == 0:
-                return _ConfigEvaluation(False, False)
+                return False
             pass_gate = False
-            exposures = []
             for gate in target:
                 other_result = self.check_gate(user, gate)
-                if other_result.unsupported:
-                    return _ConfigEvaluation(True)
+
                 new_exposure = {
                     "gate": gate,
                     "gateValue": "true" if other_result.boolean_value else "false",
                     "ruleID": other_result.rule_id
                 }
-                exposures.append(new_exposure)
-                if other_result.secondary_exposures is not None and len(
-                        other_result.secondary_exposures) > 0:
-                    exposures = exposures + other_result.secondary_exposures
+                end_result.secondary_exposures.append(new_exposure)
+
                 pass_gate = pass_gate or other_result.boolean_value if type == "MULTI_PASS_GATE" else pass_gate or not other_result.boolean_value
                 if pass_gate:
                     break
-            return _ConfigEvaluation(
-                False, pass_gate, {}, "", exposures)
+            return pass_gate
         if type == "IP_BASED":
             value = self.__get_from_user(user, field)
             if value is None:
@@ -328,7 +385,7 @@ class _Evaluator:
                 if ip is not None and field == "country":
                     value = self._country_lookup.lookupStr(ip)
             if value is None:
-                return _ConfigEvaluation(False, False)
+                return False
         elif type == "UA_BASED":
             value = self.__get_from_user_agent(user, field)
         elif type == "USER_FIELD":
@@ -346,99 +403,92 @@ class _Evaluator:
                 salt_str + "." + unit_id) % 1000)
         elif type == "UNIT_ID":
             value = self.__get_unit_id(user, id_Type)
-        else:
-            return _ConfigEvaluation(True)
 
         op = condition.get("operator")
+        user_bucket = condition.get("user_bucket")
         if op == "gt":
             val = self.__get_value_as_float(value)
             target = self.__get_value_as_float(target)
             if val is None or target is None:
-                return _ConfigEvaluation(False, False)
-            return _ConfigEvaluation(False, val > target)
+                return False
+            return val > target
         if op == "gte":
             val = self.__get_value_as_float(value)
             target = self.__get_value_as_float(target)
             if val is None or target is None:
-                return _ConfigEvaluation(False, False)
-            return _ConfigEvaluation(False, val >= target)
+                return False
+            return val >= target
         if op == "lt":
             val = self.__get_value_as_float(value)
             target = self.__get_value_as_float(target)
             if val is None or target is None:
-                return _ConfigEvaluation(False, False)
-            return _ConfigEvaluation(False, val < target)
+                return False
+            return val < target
         if op == "lte":
             val = self.__get_value_as_float(value)
             target = self.__get_value_as_float(target)
             if val is None or target is None:
-                return _ConfigEvaluation(False, False)
-            return _ConfigEvaluation(False, val <= target)
+                return False
+            return val <= target
         if op == "version_gt":
-            res = self.__version_compare_helper(
+            return self.__version_compare_helper(
                 value, target, lambda result: result > 0)
-            return _ConfigEvaluation(False, res)
         if op == "version_gte":
-            res = self.__version_compare_helper(
+            return self.__version_compare_helper(
                 value, target, lambda result: result >= 0)
-            return _ConfigEvaluation(False, res)
         if op == "version_lt":
-            res = self.__version_compare_helper(
+            return self.__version_compare_helper(
                 value, target, lambda result: result < 0)
-            return _ConfigEvaluation(False, res)
         if op == "version_lte":
-            res = self.__version_compare_helper(
+            return self.__version_compare_helper(
                 value, target, lambda result: result <= 0)
-            return _ConfigEvaluation(False, res)
         if op == "version_eq":
-            res = self.__version_compare_helper(
+            return self.__version_compare_helper(
                 value, target, lambda result: result == 0)
-            return _ConfigEvaluation(False, res)
         if op == "version_neq":
-            res = self.__version_compare_helper(
+            return self.__version_compare_helper(
                 value, target, lambda result: result != 0)
-            return _ConfigEvaluation(False, res)
         if op == "any":
-            return _ConfigEvaluation(
-                False, self.__match_string_in_array(
-                    value, target, lambda a, b: a.upper().lower() == b.upper().lower()))
+            if user_bucket is not None:
+                return self.__lookup_user_bucket(value, user_bucket)
+            return self.__find_string_in_array(
+                value, condition)
         if op == "none":
-            return _ConfigEvaluation(
-                False, not self.__match_string_in_array(
-                    value, target, lambda a, b: a.upper().lower() == b.upper().lower()))
+            if user_bucket is not None:
+                return not self.__lookup_user_bucket(value, user_bucket)
+            return not self.__find_string_in_array(
+                value, condition)
         if op == "any_case_sensitive":
-            return _ConfigEvaluation(False, self.__match_string_in_array(
-                value, target, lambda a, b: a == b))
+            return self.__find_string_in_array(
+                value, condition)
         if op == "none_case_sensitive":
-            return _ConfigEvaluation(False, not self.__match_string_in_array(
-                value, target, lambda a, b: a == b))
+            return not self.__find_string_in_array(
+                value, condition)
         if op == "str_starts_with_any":
-            return _ConfigEvaluation(False, self.__match_string_in_array(
+            return self.__match_string_in_array(
                 value, target, lambda a, b: a.upper().lower().startswith(
-                    b.upper().lower())))
+                    b.upper().lower()))
         if op == "str_ends_with_any":
-            return _ConfigEvaluation(False, self.__match_string_in_array(
+            return self.__match_string_in_array(
                 value, target, lambda a, b: a.upper().lower().endswith(
-                    b.upper().lower())))
+                    b.upper().lower()))
         if op == "str_contains_any":
-            return _ConfigEvaluation(
-                False, self.__match_string_in_array(
-                    value, target, lambda a, b: b.upper().lower() in a.upper().lower()))
+            return self.__match_string_in_array(
+                value, target, lambda a, b: b.upper().lower() in a.upper().lower())
         if op == "str_contains_none":
-            return _ConfigEvaluation(
-                False, not self.__match_string_in_array(
-                    value, target, lambda a, b: b.upper().lower() in a.upper().lower()))
+            return not self.__match_string_in_array(
+                value, target, lambda a, b: b.upper().lower() in a.upper().lower())
         if op == "str_matches":
             str_value = self.__get_value_as_string(value)
             str_target = self.__get_value_as_string(target)
             if str_value is None or str_target is None:
-                return _ConfigEvaluation(False, False)
-            return _ConfigEvaluation(False, bool(
-                re.search(str_target, str_value)))
+                return False
+            return bool(
+                re.search(str_target, str_value))
         if op == "eq":
-            return _ConfigEvaluation(False, value == target)
+            return value == target
         if op == "neq":
-            return _ConfigEvaluation(False, value != target)
+            return value != target
         if op == "before":
             return self.__compare_dates(value, target, lambda a, b: a < b)
         if op == "after":
@@ -448,10 +498,25 @@ class _Evaluator:
                 value, target, lambda a, b: a.date() == b.date())
         if op in ("in_segment_list", "not_in_segment_list"):
             in_list = self.__check_id_in_list(value, target)
-            return _ConfigEvaluation(
-                False, in_list if op == "in_segment_list" else not in_list)
+            return in_list if op == "in_segment_list" else not in_list
+        if op == "array_contains_any":
+            if not isinstance(value, list):
+                return False
+            return self.__arrays_have_common_value(value, condition)
+        if op == "array_contains_none":
+            if not isinstance(value, list):
+                return False
+            return not self.__arrays_have_common_value(value, condition)
+        if op == "array_contains_all":
+            if not isinstance(value, list):
+                return False
+            return self.__arrays_have_all_values(value, condition)
+        if op == "not_array_contains_all":
+            if not isinstance(value, list):
+                return False
+            return not self.__arrays_have_all_values(value, condition)
 
-        return _ConfigEvaluation(True)
+        return False
 
     def __get_from_user(self, user, field):
         value = None
@@ -495,9 +560,13 @@ class _Evaluator:
         return None
 
     def __compute_user_hash(self, input):
-        return unpack('>Q', sha256(str(input).encode('utf-8')).digest()[:8])[0]
+        return sha256_hash(input)
 
     def __eval_pass_percentage(self, user, rule, config):
+        if rule.get("passPercentage", 0) == 100.0:
+            return True
+        if rule.get("passPercentage", 0) == 0.0:
+            return False
         rule_salt = rule.get("salt", rule.get("id", ""))
         id = self.__get_unit_id(user, rule.get("idType", "userID")) or ""
         config_salt = config.get("salt", "")
@@ -530,6 +599,45 @@ class _Evaluator:
                 return True
         return False
 
+    def __find_string_in_array(self, value, condition):
+        str_value = self.__get_value_as_string(value)
+        target = condition.get("fast_target_value")
+        op = condition.get("operator")
+        if str_value is None or target is None:
+            return False
+        if op in ('any', 'none'):
+            return str_value.upper().lower() in target
+        return str_value in target
+
+    def __arrays_have_common_value(self, value, condition):
+        fast_target = condition.get("fast_target_value")
+        for target_val in fast_target:
+            int_target_val = self.safe_parse_int(target_val)
+            if int_target_val and int_target_val in value:
+                return True
+            if target_val in value:
+                return True
+        return False
+
+    def __arrays_have_all_values(self, value, condition):
+        fast_target = condition.get("fast_target_value")
+        for target_val in fast_target:
+            int_target_val = self.safe_parse_int(target_val)
+            if int_target_val not in value and target_val not in value:
+                return False
+        return True
+
+    def safe_parse_int(self, value):
+        try:
+            return int(value)
+        except (ValueError, TypeError):
+            return None
+
+    def __lookup_user_bucket(self, val, lookup):
+        if isinstance(val, int):
+            return val in lookup
+        return False
+
     def __version_compare(self, v1, v2, compare):
         p1 = v1.split(".")
         p2 = v2.split(".")
@@ -623,17 +731,14 @@ class _Evaluator:
 
     def __compare_dates(self, first, second, compare):
         if first is None and second is None:
-            return _ConfigEvaluation(False, False)
+            return False
 
         first_date = self.__get_date(first)
         second_date = self.__get_date(second)
         if first_date is None or second_date is None:
-            return _ConfigEvaluation(False, False)
+            return False
 
-        return _ConfigEvaluation(
-            False,
-            compare(first_date, second_date)
-        )
+        return compare(first_date, second_date)
 
     def __get_date(self, d):
         if d is None:
@@ -641,6 +746,6 @@ class _Evaluator:
 
         epoch = int(d)
         if len(str(d)) >= 11:
-            epoch /= 1000
+            epoch //= 1000
 
         return datetime.fromtimestamp(epoch)
