@@ -1,12 +1,15 @@
 from typing import Optional
-from .statsig_event import StatsigEvent
-from .statsig_user import StatsigUser
-from .statsig_server import StatsigServer
-from .statsig_options import StatsigOptions
+
+from . import globals, FeatureGate
+from .client_initialize_formatter import ClientInitializeResponse
 from .dynamic_config import DynamicConfig
 from .layer import Layer
-from .client_initialize_formatter import ClientInitializeResponse
-from . import globals
+from .statsig_event import StatsigEvent
+from .statsig_options import StatsigOptions
+from .statsig_server import StatsigServer
+from .statsig_user import StatsigUser
+from .utils import HashingAlgorithm
+from .version import __version__
 
 __instance = StatsigServer()
 
@@ -23,16 +26,32 @@ def initialize(secret_key: str, options: Optional[StatsigOptions] = None):
 
     if options.custom_logger is not None:
         globals.set_logger(options.custom_logger)
-    elif options.enable_debug_logs:
-        globals.enable_debug_logs()
+    elif options.output_logger_level is not None:
+        globals.set_log_level(options.output_logger_level)
+
+    globals.logger.info(
+        f"Initializing Statsig SDK (v{__version__}) instance. "
+        f"Current environment tier: {options.get_sdk_environment_tier()}."
+    )
 
-    globals.logger.log_process("Initialize", "Starting...")
     __instance.initialize(secret_key, options)
 
-    if __instance._initialized:
-        globals.logger.log_process("Initialize", "Done")
+    if __instance.is_initialized():
+        if __instance.is_store_populated():
+            globals.logger.info(
+                f"Statsig SDK instance initialized successfully with data from {__instance.get_init_source()}")
+        else:
+            globals.logger.error(
+                "Statsig SDK instance initialized, but config store is not populated. The SDK is using default values for evaluation.")
     else:
-        globals.logger.log_process("Initialize", "Failed")
+        globals.logger.error("Statsig SDK instance Initialized failed!")
+
+
+def is_initialized():
+    """
+    Checks if the Statsig instance has been initialized
+    """
+    return __instance.is_initialized()
 
 
 def check_gate(user: StatsigUser, gate: str) -> bool:
@@ -46,6 +65,18 @@ def check_gate(user: StatsigUser, gate: str) -> bool:
     return __instance.check_gate(user, gate)
 
 
+def get_feature_gate(user: StatsigUser, gate: str, log_exposure=True) -> FeatureGate:
+    """
+    Gets the FeatureGate object for the given user
+
+    :param user: The StatsigUser object used for the evaluation
+    :param gate: The name of the gate
+    :param log_exposure: Boolean flag to optionally disable exposure logging (Default: True)
+    :return: A FeatureGate object
+    """
+    return __instance.get_feature_gate(user, gate, log_exposure)
+
+
 def check_gate_with_exposure_logging_disabled(user: StatsigUser, gate: str) -> bool:
     """
     Checks the value of a Feature Gate for the given user without logging an exposure event
@@ -99,6 +130,15 @@ def manually_log_config_exposure(user: StatsigUser, config: str):
     __instance.manually_log_config_exposure(user, config)
 
 
+def log_exposure_for_config(config_value: DynamicConfig):
+    """
+    Logs an exposure event using a DynamicConfig object
+
+    :param config_value: The DynamicConfig object to log exposure for
+    """
+    __instance.log_exposure_for_config(config_value)
+
+
 def get_experiment(user: StatsigUser, experiment: str) -> DynamicConfig:
     """
     Gets the DynamicConfig value of an Experiment for the given user
@@ -266,7 +306,9 @@ def remove_all_overrides():
     __instance.remove_all_overrides()
 
 
-def get_client_initialize_response(user: StatsigUser, client_sdk_key: Optional[str] = None) -> ClientInitializeResponse:
+def get_client_initialize_response(user: StatsigUser, client_sdk_key: Optional[str] = None,
+                                   hash: Optional[HashingAlgorithm] = HashingAlgorithm.SHA256,
+                                   include_local_overrides: Optional[bool] = False) -> ClientInitializeResponse:
     """
     Gets all evaluated values for the given user.
     These values can then be given to a Statsig Client SDK via bootstrapping.
@@ -276,7 +318,7 @@ def get_client_initialize_response(user: StatsigUser, client_sdk_key: Optional[s
     :param client_sdk_key: (Optional) The client sdk key to use for bootstrapping
     :return: An initialize response containing evaluated gates/configs/layers
     """
-    return __instance.get_client_initialize_response(user, client_sdk_key)
+    return __instance.get_client_initialize_response(user, client_sdk_key, hash, include_local_overrides)
 
 
 def evaluate_all(user: StatsigUser):
